#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2014 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.IntegerVectorEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.Orienteering {
  public sealed class BestOrienteeringSolutionAnalyser : SingleSuccessorOperator, IAnalyzer {
    public bool EnabledByDefault {
      get { return true; }
    }

    public IScopeTreeLookupParameter<IntegerVector> IntegerVector {
      get { return (IScopeTreeLookupParameter<IntegerVector>)Parameters["IntegerVector"]; }
    }
    public LookupParameter<DoubleMatrix> CoordinatesParameter {
      get { return (LookupParameter<DoubleMatrix>)Parameters["Coordinates"]; }
    }
    public ILookupParameter<IntValue> StartingPointParameter {
      get { return (ILookupParameter<IntValue>)Parameters["StartingPoint"]; }
    }
    public ILookupParameter<IntValue> TerminalPointParameter {
      get { return (ILookupParameter<IntValue>)Parameters["TerminalPoint"]; }
    }
    public ILookupParameter<DoubleArray> ScoresParameter {
      get { return (ILookupParameter<DoubleArray>)Parameters["Scores"]; }
    }

    public ScopeTreeLookupParameter<DoubleValue> QualityParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ScopeTreeLookupParameter<DoubleValue> PenaltyParameter {
      get { return (ScopeTreeLookupParameter<DoubleValue>)Parameters["Penalty"]; }
    }
    public LookupParameter<OrienteeringSolution> BestSolutionParameter {
      get { return (LookupParameter<OrienteeringSolution>)Parameters["BestSolution"]; }
    }
    public ValueLookupParameter<ResultCollection> ResultsParameter {
      get { return (ValueLookupParameter<ResultCollection>)Parameters["Results"]; }
    }
    public LookupParameter<DoubleValue> BestKnownQualityParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BestKnownQuality"]; }
    }
    public LookupParameter<IntegerVector> BestKnownSolutionParameter {
      get { return (LookupParameter<IntegerVector>)Parameters["BestKnownSolution"]; }
    }
    public LookupParameter<DoubleValue> BestKnownSolutionPenaltyParameter {
      get { return (LookupParameter<DoubleValue>)Parameters["BestKnownSolutionPenalty"]; }
    }

    [StorableConstructor]
    private BestOrienteeringSolutionAnalyser(bool deserializing) : base(deserializing) { }
    private BestOrienteeringSolutionAnalyser(BestOrienteeringSolutionAnalyser original, Cloner cloner) : base(original, cloner) { }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new BestOrienteeringSolutionAnalyser(this, cloner);
    }
    public BestOrienteeringSolutionAnalyser()
      : base() {
      Parameters.Add(new ScopeTreeLookupParameter<IntegerVector>("IntegerVector", "The Orienteering solutions which should be analysed."));
      Parameters.Add(new LookupParameter<DoubleMatrix>("Coordinates", "The x- and y-Coordinates of the points."));
      Parameters.Add(new LookupParameter<IntValue>("StartingPoint", "Index of the starting point."));
      Parameters.Add(new LookupParameter<IntValue>("TerminalPoint", "Index of the ending point."));
      Parameters.Add(new LookupParameter<DoubleArray>("Scores", "The scores of the points."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Quality", "The qualities of the Orienteering solutions which should be analyzed."));
      Parameters.Add(new ScopeTreeLookupParameter<DoubleValue>("Penalty", "The applied penalty of the Orienteering solutions."));
      Parameters.Add(new LookupParameter<OrienteeringSolution>("BestSolution", "The best Orienteering solution."));
      Parameters.Add(new ValueLookupParameter<ResultCollection>("Results", "The result collection where the best Orienteering solution should be stored."));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownQuality", "The quality of the best known solution of this Orienteering instance."));
      Parameters.Add(new LookupParameter<IntegerVector>("BestKnownSolution", "The best known solution of this Orienteering instance."));
      Parameters.Add(new LookupParameter<DoubleValue>("BestKnownSolutionPenalty", "The penalty of the best known solution of this Orienteering instance."));
    }

    public override IOperation Apply() {
      var solutions = IntegerVector.ActualValue;
      var qualities = QualityParameter.ActualValue;
      var penalties = PenaltyParameter.ActualValue;
      var results = ResultsParameter.ActualValue;
      var bestKnownQuality = BestKnownQualityParameter.ActualValue;

      int bestIndex = qualities.Select((quality, index) => new { index, quality.Value }).OrderByDescending(x => x.Value).First().index;

      if (bestKnownQuality == null || qualities[bestIndex].Value > bestKnownQuality.Value) {
        BestKnownQualityParameter.ActualValue = new DoubleValue(qualities[bestIndex].Value);
        BestKnownSolutionParameter.ActualValue = (IntegerVector)solutions[bestIndex].Clone();
        BestKnownSolutionPenaltyParameter.ActualValue = new DoubleValue(penalties[bestIndex].Value);
      }

      var solution = BestSolutionParameter.ActualValue;
      var coordinates = CoordinatesParameter.ActualValue;
      var startingPoint = StartingPointParameter.ActualValue;
      var terminalPoint = TerminalPointParameter.ActualValue;
      var scores = ScoresParameter.ActualValue;
      if (solution == null) {
        solution = new OrienteeringSolution(
          (IntegerVector)solutions[bestIndex].Clone(),
          coordinates,
          startingPoint,
          terminalPoint,
          scores,
          new DoubleValue(qualities[bestIndex].Value),
          new DoubleValue(penalties[bestIndex].Value));
        BestSolutionParameter.ActualValue = solution;
        results.Add(new Result("Best Orienteering Solution", solution));
        results.Add(new Result("BestKnownSolutionPenalty", solution.Penalty));
      } else {
        if (solution.Quality.Value < qualities[bestIndex].Value) {
          solution.Coordinates = coordinates;
          solution.Scores = scores;
          solution.IntegerVector = (IntegerVector)solutions[bestIndex].Clone();
          solution.Quality.Value = qualities[bestIndex].Value;
          solution.Penalty.Value = penalties[bestIndex].Value;
          results["BestKnownSolutionPenalty"].Value = solution.Penalty;
        }
      }

      return base.Apply();
    }
  }
}